(require (lib "list.ss"))
(require (lib "pretty.ss"))
(load "match.scm")
(load "production.scm")
(load "zookeeper.scm")
(load "goaltree.scm")

;; A helper function that tells you if an expression has any unknown variables
;; in it.
(define (has-variables? exp)
  (cond [(variable? exp) #t]
        [(not (pair? exp)) #f]
        [else (or (has-variables? (car exp)) (has-variables? (cdr exp)))]))

;; This function, which you need to write, takes in a hypothesis that can be determined
;; using a set of rules, and outputs a goal tree of which statements it would need to
;; test to prove that hypothesis. Refer to the problem set (section 2.1) for more
;; detailed specifications and examples.
(define (backchain-to-goal-tree rules hypothesis)
  (simplify (cons 'OR (trigger (match-rules rules hypothesis rules) hypothesis)))
)

;; If there were no matches, return the consequent, otherwise return the matches
(define (trigger matches hypothesis) (if (null? matches) (list hypothesis) matches))

;; Traverses the list of rules, if there is a match it processes the antecedents in an AND node
;; otherwise it returns nothing (allowing trigger to use the consequent)
(define (match-rules rules hypothesis rules-list) 
   (let ((matches (next-match rules hypothesis)))
    (if (pair? matches)
        ;We found a match
        (cons 
              (process-rule (car matches) hypothesis rules-list)
              (match-rules (cdr matches) hypothesis rules-list)
        )
        ;Otherwise return the empty list
        '()
    )
   )
)

;; Finds the next rule which matches, returns that rule + the remaining rules
(define (next-match rules hypothesis)
  (if (null? rules) 
      '()
      (if (match (rule-consequent (car rules)) hypothesis)
          rules
          (next-match (cdr rules) hypothesis)
      )
  )
)
  
;; Processes a matching rule and consequent
(define (process-rule rule hypothesis rules)
  (process-antecedents (substitute (rule-antecedent rule) (match (rule-consequent rule) hypothesis)) rules)
)

;; Processes the antecedents of a rule, provided that all variables have been
;; substituted
(define (process-antecedents antecedents rules)
 (cond 
     ;The first element of this antecedent is an 'AND or 'OR
     ((or (eqv? 'OR (car antecedents)) (eqv? 'AND (car antecedents)))
       (cons (car antecedents) (process-antecedents (cdr antecedents) rules)))
     ;If the first element is atomic, then there was only one antecedent
     ((atom? (car antecedents))
       (backchain-to-goal-tree rules antecedents))
     ;Otherwise, it must be one of several, so process it and the rest
     (else 
        (if (pair? (cdr antecedents))
         (cons (backchain-to-goal-tree rules (car antecedents)) (process-antecedents (cdr antecedents) rules))
         (cons (backchain-to-goal-tree rules (car antecedents)) '())
        )
     )
 )
)

;; Here's an example of running the backward chainer - uncomment it to see it work:
 (pretty-print (backchain-to-goal-tree zoo-rules '(opus is a penguin)))
 
;(AND (OR (opus has feathers) (AND (OR (opus flies)) (OR (opus lays eggs)))) (OR (opus does not fly)) (OR (opus swims)) (OR (opus has black and white color))) 
 
