; This game contains the game logic and helper functions for the eight-pawn game.
; You'll want to look at this code, but don't change it -- only the code you write
; in ps3.scm will be graded.

(require (lib "list.ss"))
(require (lib "pretty.ss"))

; A _position_ is a list of three things:
;   1. the tag 'position
;   2. the color that moves next, W or B
;   3. the _board_
;
; A _board_ is a list of _pieces_, indicating where all the pieces in play are.
;
; A _piece_ is a list of the piece's color, the column the piece is in (1-4),
; and the row the piece is in (1-4).
;
; Within the code that follows, spaces on the board are expressed as a piece,
; regardless of whether that piece actually exists. So the bottom left corner
; of the board is passed to functions as (W 1 1), or maybe as (B 1 1). The
; piece's color is disregarded if all that matters is the location.
;
; Here's the starting position of the game:
(define startpos
  '(position W ((W 1 1) (W 2 1) (W 3 1) (W 4 1) (B 1 4) (B 2 4) (B 3 4) (B 4 4))))

(define (opposite-color color)
  (if (eq? color 'W) 'B 'W))

; Accessors for positions
(define get-color cadr)
(define get-board caddr)

; Accessors for minimax results
(define get-move car)
(define get-score cadr)

; One possible evaluation function is the number of moves the current
; player can make. Consider the expected value to be 3. So if Black can
; make 2 moves, this evaluation gives -1 for Black and 1 for White.
;
; As it may be useful in the heuristic, the heuristic function gets the
; number of levels left to search as an argument.
(define (simple-evaluate pos player depth)
  (let* ([board (get-board pos)]
         [color-to-move (get-color pos)]
         [score (if (check-loss pos) -1000
                   (- (length (valid-moves pos)) 3))])

    ; Negate the score if the position is for the other player.
    (if (eq? player color-to-move) score (- score))))

; Given two (position score) results from the evaluation function, return
; the one with the higher score.
(define (maximize-position p1 p2)
  (if (> (get-score p1) (get-score p2)) p1 p2))

; Given two (position score) results from the evaluation function, return
; the one with the lower score.
(define (minimize-position p1 p2)
  (if (< (get-score p1) (get-score p2)) p1 p2))

; The main minimax search procedure. Takes 5 arguments:
;  position: the current position
;  depth: how many _more_ levels to search from here
;  player: which player we're evaluating for. We could be looking at a
;          position where it's White's turn from Black's point of view.
;  max?: are we maximizing or minimizing at this step?
;  evaluation-function: a static evaluation function, which takes in
;                       a position and returns a number for how good it is.
;
; It returns a list of two items:
;  * the position to move to
;  * the heuristic value of that position

(define (minimax-search position depth player max? evaluation-function)
  (if (or (= depth 0) (check-loss position))
      ; Stop expanding the game tree here, either because someone won or
      ; we've expanded far enough.
      ; We're not making any moves from here, so return #f as the move.
      (list #f (evaluation-function position player depth))
      
      ; Otherwise, find the valid moves, recursively evaluate them,
      ; and return the one with the highest/lowest score.
      
      (foldl (if max? maximize-position minimize-position)
             (list #f (if max? -2000 2000)) ; start with an extreme value that won't actually be returned
             
             ; Expand every position we can get to using minimax-search, and attach to each one
             ; the position we move to so that we know which move to make.
             ;
             ; The cadr of the result obtained from minimax-search is the heuristic value.
             ; We disregard the car, which is the move to make from there - we only need
             ; to make one move at a time.
             (map (lambda (newpos)
                    (list newpos (cadr (minimax-search newpos (- depth 1) player
                                                       (not max?) evaluation-function))))
                  (valid-moves position)))))

; A computerized player that searches using minimax to a certain depth.
; You need to provide it with an evaluation function; simple-evaluate will work.
(define (minimax-player depth evaluation-function)
  (lambda (position)
    (let* ([analysis (minimax-search position depth (cadr position) #t evaluation-function)]
           [best-move (car analysis)]
           [score (cadr analysis)])
      (car analysis))))

; A computerized player that outputs the board as it goes, and
; comments on how it's doing.
(define (chatty-minimax-player depth evaluation-function)
  (lambda (position)
    (display-board position)
    (newline)
    (let* ([analysis (minimax-search position depth (cadr position) #t evaluation-function)]
           [best-move (car analysis)]
           [score (cadr analysis)])
      (display "Minimax value: ") (display score) (newline)
      (cond [(>= score 1000) (display "Hee hee. I'm going to win.") (newline)]
            [(<= score -1000) (display "I'm doomed.") (newline)])
      (car analysis))))

; This computerized player won't work until you write alpha-beta-search.
; The last two parameters are alpha and beta.
(define (alpha-beta-player depth evaluation-function)
  (lambda (position)
    (let* ([analysis (alpha-beta-search position depth (cadr position) #t evaluation-function -2000 2000)]
           [best-move (car analysis)]
           [score (cadr analysis)])
      (car analysis))))

; Same with this player.
(define (chatty-alpha-beta-player depth evaluation-function)
  (lambda (position)
    (display-board position)
    (newline)
    (let* ([analysis (alpha-beta-search position depth (cadr position) #t evaluation-function -2000 2000)]
           [best-move (car analysis)]
           [score (cadr analysis)])
      (display "Minimax value: ") (display score) (newline)
      (cond [(>= score 1000) (display "Hee hee. I'm going to win.") (newline)]
            [(<= score -1000) (display "I'm doomed.") (newline)])
      (car analysis))))

; Determine if a space is occupied. If so, return the color of the piece occupying it.
; If not, return #f.
(define (occupied board space)
  (if (null? board) #f
      (if (equal? (cdr space) (cdar board))
          (caar board)
          (occupied (cdr board) space))))

; Clear a space on the board (probably because a piece there is being captured).
(define (clear-space board space)
  (filter (lambda (s) (not (equal? (cdr s) (cdr space)))) board))

; Is this space on the board at all?
(define (valid-space space)
  (and (> (cadr space) 0)
       (< (cadr space) 5)
       (> (caddr space) 0)
       (< (caddr space) 5)))

; A list of all positions that can be moved to from this position.
(define (valid-moves position)
  (let ([board (get-board position)]
        [color (get-color position)])
    (define (valid-moves-helper remain)
      (if (null? remain) ()
          (let ([current (car remain)])
            (if (not (eq? (car current) color))
                (valid-moves-helper (cdr remain))
                (let* ([forward-offset (if (eq? color 'W) 1 -1)]
                       [other-color (opposite-color color)]
                       [forward (list color (cadr current) (+ forward-offset (caddr current)))]
                       [left-capture (list color (- (cadr current) 1) (+ forward-offset (caddr current)))]
                       [right-capture (list color (+ (cadr current) 1) (+ forward-offset (caddr current)))]
                       [remove-this-piece (remove current board)])
                  (append
                   (filter (lambda (x) x)
                           (list
                            (if (and (valid-space forward) (not (occupied board forward)))
                                (list 'position other-color (cons forward remove-this-piece))
                                #f)
                            (if (and (valid-space left-capture) (eq? (occupied board left-capture) other-color))
                                (list 'position other-color
                                      (cons left-capture (clear-space remove-this-piece left-capture)))
                                #f)
                            (if (and (valid-space right-capture) (eq? (occupied board right-capture) other-color))
                                (list 'position other-color
                                      (cons right-capture (clear-space remove-this-piece right-capture)))
                                #f)))
                   (valid-moves-helper (cdr remain))))))))
    (valid-moves-helper board)))

; What rank (row) is this color trying to move a pawn to?
(define (goal-rank color)
  (if (eq? color 'W) 4 1))

; Is this position a loss for the current player?
(define (check-loss position)
  (let ([board (get-board position)]
        [color (get-color position)])
    (or (not (null? (filter (lambda (piece)
                              (and (eq? (car piece) (opposite-color color))
                                   (= (caddr piece) (goal-rank (opposite-color color)))))
                            board)))
        (null? (valid-moves position)))))

; Using this function, we can count the number of static evaluations we make.
(define (make-counter proc)
  (let ([count 0])
    (lambda args
      (if (equal? args '(count)) count
          (begin
            (set! count (+ count 1))
            (apply proc args))))))

; Convert the board from a list of pieces to a list of lists indicating
; the contents of each space.
(define (board-list pos)
  (let ([ranks '(1 2 3 4)]
        [board (caddr pos)])
    (map (lambda (rank)
           (map (lambda (file)
                  (let ([piece (occupied board (list 'W file rank))])
                    (if (not piece) '- piece)))
                ranks))
         ranks)))

; Display the board to the user.
(define (display-board pos)
  (let ([board (board-list pos)])
    (map (lambda (n) (display (list-ref board n)) (newline))
         '(3 2 1 0))))

; A player procedure that asks the human what to do at each step.
(define (human-player)
  (lambda (position)
    (display-board position)
    (newline)
    (let ([moves (valid-moves position)])
      (let loop ([remain moves]
                 [count 1])
        (if (null? remain) ()
            (let ([one-pos (car remain)])
              (display count) (display ": ")
              (display (caaddr one-pos))
              (newline)
              (loop (cdr remain) (+ count 1)))))
      (newline)
      (display "Which move do you want to make? ")
      (let input-loop ([move-index (read)])
        (if (or (not (number? move-index))
                (< move-index 1)
                (> move-index (length moves)))
            (begin
              (display "That's not a choice. Please type one of the numbers given.")
              (newline)
              (input-loop (read)))
            (let ([newpos (list-ref moves (- move-index 1))])
              newpos))))))

; This procedure lets you play the eight-pawn game, given two "player"
; procedures that take in a position and return a position to move to.
(define (play-game position player1 player2)
  (if (check-loss position)
      (begin
        (display-board position)
        (display "The winner is: ")
        (opposite-color (cadr position)))
      (play-game (player1 position) player2 player1)))
