(require (lib "list.ss"))

;; Gets the features of an example
(define get-features car)

;; Gets the class of an example
(define get-class cdr)

;; Uses the cos-theta function to determine an angle-based metric
(define (angle-metric x y)
  (- 1 (cos-theta x y)))

;; Returns the log (base 2) of a number
(define (log2 n)
  (/ (log n) (log 2)))

;; Gets the k examples nearest a given feature set
(define (get-k-nearest k distance-metric features examples)
  (if (>= k (length examples))
      examples
      (let ([sorted-examples (sort examples
                                   (lambda (a b)
                                     (< (distance-metric features (get-features a))
                                        (distance-metric features (get-features b)))))])                     
        (list-tail (reverse sorted-examples) (- (length sorted-examples) k)))))


;; Tests a classifier against a set of labeled data, indicating the
;; number of correct classifications
(define (validate-classifier classifier examples)
  (let* ((classifier-outputs (map (lambda (example)
				    (classifier (get-features example)))
				  examples))
	 (num-examples (length examples))
	 (num-correct (foldr + 0 (map (lambda (class-output example)
					     (if (equal? class-output
							 (cdr example))
						 1
						 0)
					     )
					   classifier-outputs
					   examples
					   )
				  )
		      )
	 )
    (cons (list num-examples num-correct (exact->inexact (/ num-correct num-examples)))
	  classifier-outputs)
    )
  )

;; Checks whether or not an example is valid
;; Examples should be a pair of features and classifications
(define (valid-example? x)
  (and (pair? x) (string? (cdr x))))

;; TODO: what is this for?
;(define (and-func list)
;  (if (null? list)
;      #t
;      (and (car list)
;	   (and-func (cdr list))
;	   )
;      )
;  )

;; Converts the data in a list of examples into vectors
(define (dataset-list-to-vector-examples lst)
  (map (lambda (elt)
	 (cons (list->vector (car elt))
	       (cdr elt))
	 )
       lst)
  )

;; Extracts only the class names from a list of examples
(define (classes-from-examples examples)
  (map cdr examples))

;; Extracts only the features from a list of exampls
(define (features-from-examples examples)
  (map car examples))

;; Gets the min and max coordinates in a list of vectors
;; TODO: what is this for?
(define (min-max-coords veclist)
  (let ((min-x -1)
	(max-x 1)
	(min-y -1)
	(max-y 1)
	)
    (map (lambda (vec)
	   (let ((x (vector-ref vec 0))
		 (y (vector-ref vec 1))
		 )
	     (if (< x min-x)
		 (set! min-x x))
	     (if (> x max-x)
		 (set! max-x x))
	     (if (< y min-y)
		 (set! min-y y))
	     (if (> y max-y)
		 (set! max-y y))
	     )
	   )
	 veclist
	 )
    (list min-x max-x min-y max-y)
    )
  )

;; TODO: what does this do?
(define (plot-points list-of-vecs list-of-classes classes-color-alist)
  (map (lambda (vec)
	 (if (not (r2-vector? vec))
	     (error "Non R2 vector supplied. What are you trying to pull?\n" vec)
	     )
	 )
       list-of-vecs
       )
  (let* ((min-max (min-max-coords list-of-vecs))
	 (min-x (* 1.2 (first min-max)))
	 (max-x (* 1.2 (second min-max)))
	 (min-y (* 1.2 (third min-max)))
	 (max-y (* 1.2 (fourth min-max)))
;;	 (win (frame min-x min-y max-x max-y))
	 (win (make-display-frame min-x max-x min-y max-y
				  300 300))
	 (blob-radius (/ (abs (- max-x min-x)) 30.0))
	 )
;;    (pp (list 'Window-limits min-max))
    (let loop ((vecs list-of-vecs)
	       (classes list-of-classes)
	       )
      (if (null? vecs)
	  (cons win (list min-x max-x min-y max-y))
	  (let* ((cur-vec (car vecs))
		 (cur-class (car classes))
		 (cur-color (assoc cur-class classes-color-alist))
		 )
	    (if (null? cur-color)
		(error "No color supplied for class. Stop that!\n" cur-class)
		(set! cur-color (cadr cur-color))
		)
	    (pp (list cur-vec cur-class cur-color blob-radius))
	    (graphics-operation win 'set-foreground-color cur-color)
	    (graphics-operation win 'fill-circle
				(vector-ref cur-vec 0) (vector-ref cur-vec 1)
				blob-radius)
	    (loop (cdr vecs) (cdr classes))
	    )
	  )
      )
    )
  )

;; Checks whether or not x is a list consisting only of these symbols: [yes no maybe]
(define (is-yes-no-maybe-list? x)
  (if (not (list? x))
      #f
      (foldr (lambda (x y) (and x y))
		  #t
		  (map (lambda (v) (member v '(yes no maybe)))
		       x))))

;; Converts a yes/no/maybe vote into a numerical value (yes:1, no:-1, maybe:0)
(define (ynm-to-num val)
  (cond [(eq? val 'yes) 3]
	[(eq? val 'no) 1]
	[(eq? val 'maybe) 2]))

;; Converts an entire dataset from yes/no/maybe votes into numerical vectors
(define (congress-to-metric data)
  (map (lambda (example)
	 (cons (map ynm-to-num (get-features example))
	       (get-class example)))
       data))

;; Converts an entirer dataset from yes/no/maybe votes into inputs suitable
;;  for a back-propagating neural net
(define (congress-to-bpnn-metric data)
  (map (lambda (example)
	 (cons (list->vector (map ynm-to-num (get-features example)))
	       (make-vector 1
			    (if (equal? (get-class example) "republican")
				1.0
				-1.0))))
       data))