(load "data.scm")

; Creates a simple base classifier, which checks if a particular attribute
; has a particular value. "Polarity" is either 1 or -1: if polarity is 1,
; then this classifier will output 1 if the attribute matches and -1 otherwise.
; If polarity is -1, it's the other way around.
;
; A classifier accepts the special input 'describe, in which case it will
; display the information it uses to make a classification.
(define (make-base-classifier attribute value polarity)
  (lambda (x)
    (if (eq? x 'describe)
        (list (attribute-name attribute) value polarity)
        (if (equal? ((attribute-extractor attribute) x) value)
            polarity
            (- polarity)))))

; Evaluate the error rate of a classifier, taking the weights of the data
; into account.
(define (classifier-error-rate classifier weighted-data)
  (define (helper data-left total-error)
    (if (null? data-left) total-error
        (let* ([datum (car data-left)]
               [weight (get-weight datum)]
               [info (get-info datum)]
               [actual (get-class datum)])
          (helper (cdr data-left)
                  (+ total-error
                     (if (= (classifier info) actual)
                         0
                         weight))))))
  (exact->inexact (/ (helper weighted-data 0)
                     (foldl + 0 (map car weighted-data)))))

; Return a modified version of a weighted datum, whose weight
; has been multiplied by factor.
(define (reweight-datum datum factor)
  (let ([weight (get-weight datum)]
        [info (get-info datum)]
        [actual (get-class datum)])
    (make-datum (* factor weight) info actual)))

; Given a set of data and a classifier, reweight the data so that correctly
; classified data get less weight and incorrectly classified data get more.
(define (reweight-data classifier weighted-data error-rate)
    (normalize
     (let loop ((data weighted-data))
     (cond ((pair? data) 
           ;Place a reweighted datum by the given formula
           (cons (reweight-datum (car data) (if (equal? (get-class (car data)) (classifier (get-info (car data))))
                                          (/ .5 (- 1 error-rate))
                                          (/ .5 error-rate)))
                 (loop (cdr data))))
           (else '())
     )
    )
   )
)

; The weights that come from reweight-data should add up to 1. Use this
; function to accomplish that.
(define (normalize data)
  (let ([sum (foldl + 0 (map get-weight data))])
    (map (lambda (datum)
           (reweight-datum datum (/ 1 sum)))
         data)))


; Given an error rate, convert it to an alpha factor (used to assign a weight
; to classifiers).
(define (error-to-alpha error-rate)
  (* 0.5 (log (/ (- 1 error-rate) error-rate))))

; Compare two classifiers over a body of data, and return the better one.
(define (better-classifier c1 c2 weighted-data)
  (if (not c1) c2
      (if (not c2) c1
          (let ([c1-error (classifier-error-rate c1 weighted-data)]
                [c2-error (classifier-error-rate c2 weighted-data)])
            (if (<= c1-error c2-error) c1 c2)))))
  
; Given a procedure that produces classifiers, find the best one it can produce.
; Attributes earlier in the list take precedence; then earlier values for the same
; attribute; and for the same attribute and value, a classification of -1 takes
; precedence over a classification of 1.
(define (find-best-classifier classifier-maker attributes weighted-data)
  (let ([possibilities
         (foldl append '()
                (foldl append '()
                       (map (lambda (attr)
                              (map (lambda (val) (list (classifier-maker attr val 1)
                                                       (classifier-maker attr val -1)))
                                   (attribute-values attr)))
                            attributes)))])
    (foldl (lambda (x y) (better-classifier x y weighted-data))
           #f possibilities)))

; The sign function is 1 for a positive input and -1 for a negative input.
(define (sign x)
  (cond [(> x 0) 1]
        [(< x 0) -1]
        [else 0]))

; Given a list of lists, whose first element is a weight and whose second element is
; a classifier, output a description of each classifier along with its weight.
(define (describe-classifiers weighted-classifiers)
  (map (lambda (weighted-classifier)
         (list (car weighted-classifier)
               ((cadr weighted-classifier) 'describe)))
       weighted-classifiers))

; Returns a boost classifier, made of a weighted combination of T base
; classifiers. T is given by the input "rounds".
;
; This should use classifier-maker to find the classifier that best fits
; the weighted data at each round, and update the weighted data accordingly.
; The resulting classifiers are themselves weighted according to the "alpha"
; formula.
;
; This procedure should return a procedure that uses the weighted classifiers
; to classify an unknown datum (just the "info" part of the datum, in fact)
; as either 1 or -1. It will be tested over the political data in data.scm.
; When trained on train-data-large for 15 rounds, it should achieve an error
; rate of 3% on test-data-large.
;
; The procedure should also act differently when given two special inputs:
;   Given 'describe, it should output a list of weights and descriptions for
;   each of its base classifiers. (Use our describe-classifiers function.)
;
;   Given 'data, it should return the state of weighted-data after the last
;   round of training.
(define (make-boost-classifier weighted-data classifier-maker attributes rounds)
  (let ((training (train-boost-classifier weighted-data classifier-maker attributes rounds)))
    (let ((weighted-classifiers (omit-last training)) (data (last-pair training)))
      (lambda (x)
        (cond 
          ((eq? x 'describe) (describe-classifiers weighted-classifiers))
          ((eq? x 'data) (car data))
          ;Classification is the sign of the sum of each classifier's vote, multiplied by its weight
          (else
           (sign (foldl (lambda (x1 x2) (* (car x1) ((car (cdr x1)) x))) (list 0 (lambda (x) (0))) weighted-classifiers))
          )
        )
      )
    )
  )
)

;This function trains the boost classifier and returns the list of weighted classifiers
(define (train-boost-classifier weighted-data classifier-maker attributes rounds)
  (let loop((rounds-left rounds) (data weighted-data) (best (find-best-classifier classifier-maker attributes weighted-data)))
    (let ((error (classifier-error-rate best data)))
      (cons ;Begin creating the list of weighted-classifiers 
       (list (error-to-alpha error) best) ;Create the first weighted classifier
       (let ((new-data (reweight-data best data error)))
         (if (> rounds-left 1) (loop (- rounds-left 1) new-data (find-best-classifier classifier-maker attributes new-data)) (list new-data))
       )
      )
    )
  )
)  

;Returns everything except the last element
(define (omit-last x)
  (if (and (pair? x) (pair? (cdr x))) (cons (car x) (omit-last (cdr x))) '())
)
  
;(let ((boost (make-boost-classifier train-data-large make-base-classifier congress-attribute-specs 15)))

;(classifier-error-rate boost test-data-large)

;)  
  
; Given a classifier that responds to the 'data message (like your boost
; classifier should), return the example in its training data that was hardest
; to classify (measuring "hard to classify" the same way that the boosting
; algorithm itself does). In case of a tie, return the one earliest in the data.
;
; You should return the result as a weighted datum, including a weight, a list
; of attribute values, and the correct classification.
(define (hardest-example classifier)
  (foldl (lambda (x1 x2) (if (> (get-weight x1) (get-weight x2)) x1 x2)) 
         (make-datum 0 '() '())
         (classifier 'data))
)

;(define boost (make-boost-classifier train-data-large make-base-classifier
;                                     congress-attribute-specs 15))
;(classifier-error-rate boost test-data-large)
